class SingletonMeta(type):
    """
    A meta singleton class for helping create singleton class in python

    class Singleton(metaclass=SingletonMeta):
        pass
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]



import re
from difflib import get_close_matches
from typing import Iterable, Collection
from uuid import uuid4, UUID


def GUID() -> UUID:
    return uuid4()


def strGUID() -> str:
    return str(uuid4())


def strToGUID(strGuid: str):
    return UUID(f"{strGuid}")


# TODO refactor the method
def toUniqueList(items: Iterable[str]) -> Collection[str]:
    uniqueList = []
    illegal = -1
    patterns = [r"\d", "[a-z]", "[A-Z]", r"\W", r"[^\d|a-z|A-Z|\s|\W]"]

    for item in items:
        if isinstance(item, str) and not item.strip():
            continue

        words = re.split(r"\s", item)
        for word in words:
            previous = None
            splitIndexes = []
            for i, c in enumerate(word):
                for p in patterns:
                    if re.search(p, c) and previous != p:
                        splitIndexes.append(i)
                        previous = p
                        break

            length = len(splitIndexes)
            for i, v in enumerate(splitIndexes):
                j = i + 1
                if j < length:
                    if v + 1 == splitIndexes[j]:
                        splitIndexes[j] = illegal

            if length - splitIndexes.count(illegal) < 2:
                frag = word.lower()
                if frag not in uniqueList:
                    uniqueList.append(frag)

                continue

            i = 0
            for j in splitIndexes[1:]:
                if j == illegal:
                    continue
                frag = word[i:j].lower()
                if frag not in uniqueList:
                    uniqueList.append(frag)
                i = j

            frag = word[i:].lower()
            if frag not in uniqueList:
                uniqueList.append(frag)

    return uniqueList


def matchScore(query: str, descriptions: Iterable[str]) -> float:
    if not (descriptions and query):
        return 0.0

    unique = toUniqueList(descriptions)
    if unique:
        words = toUniqueList(re.split(r"\s", query))
        matchCount = 0
        for word in words:
            matches = get_close_matches(word, unique)
            matchCount += len(matches)

        return matchCount / len(unique)
    else:
        return 0.0


def isIEqual(str0: str, str1: str) -> bool:
    """
    Test two strings are ignored case equaled
    """
    if isinstance(str1, str) and isinstance(str0, str):
        return str0.lower() == str1.lower()
    else:
        return False



from abc import ABC, abstractmethod
from typing import Optional

from autoinsight.common.Utils import matchScore, isIEqual
from autoinsight.ident.AutomationTyping import AutomationInstance, ElementsInfo
from autoinsight.ident.IdentObjectBase import IdentObjectBase
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.services.IoCService import IoCService


class ContextBase(IdentObjectBase, ABC):
    def __init__(self,
                 ioc: IoCService = IoCService(),
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._ioc: IoCService = ioc
        self._cms: ContextManagementService = ioc.getService(ContextManagementService)
        self._cms.registerContext(self)

    def __enter__(self):
        self.setCurrent()

    def __exit__(self, *args):
        self.tearDown()

    def setCurrent(self):
        self._cms.currentContext = self

    def tearDown(self):
        self._cms.unregisterContext(self)

    @abstractmethod
    def wait(self, timeoutSeconds: int = 0):
        """
        Wait for the object to be displayed
        """
        pass

    # TODO match should also consider the control type!
    def find(self, query: str, target: IdentObjectBase = None, *args, **kwargs) -> Optional[AutomationInstance]:
        if self._automationInstance:
            try:
                info: ElementsInfo = self._automationInstance.get_elements_info()

                if target:
                    invalidIndexes = (i for i, c in enumerate(info.allCtrl) if not isIEqual(c.friendlyclassname,
                                                                                            target.classname))
                    for index in invalidIndexes:
                        if index in info.allCtrlIndexNameMaps:
                            del info.allCtrlIndexNameMaps[index]

                bestIndex, bestScore = -1, -1
                for index, names in info.allCtrlIndexNameMaps.items():
                    score = matchScore(query, names)
                    if score > bestScore:
                        bestIndex = index
                        bestScore = score

                if bestIndex > -1:
                    return info.allCtrl[bestIndex]
            except:
                return



from abc import abstractmethod
from typing import Iterable

from .ServiceBase import ServiceBase
from autoinsight.common.models.Knowledge import Knowledge
from autoinsight.ident.AutomationTyping import AutomationInstance


class KnowledgeServiceBase(ServiceBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._knowledge: Iterable[Knowledge] = []

    @abstractmethod
    def recognize(self, cmdline: str) -> AutomationInstance:
        pass

    @property
    def knowledge(self):
        return self._knowledge



import platform
from typing import Type, Optional

from autoinsight.common.EnumTypes import OSTypes
from .ConfigurationService import ConfigurationService
from .ContextManagementService import ContextManagementService
from .KnowledgeServiceBase import KnowledgeServiceBase
from .OCRServiceBase import OCRServiceBase
from .ServiceBase import ServiceBase
from .TesseractOCRService import TesseractOCRService


class IoCService(ServiceBase):
    OSType = OSTypes.fromStr(platform.system())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registeredServices = {
            OSTypes.Any: {
                OCRServiceBase: TesseractOCRService(),
                ConfigurationService: ConfigurationService(),
                ContextManagementService: ContextManagementService()
            },
            OSTypes.Windows: {},
            OSTypes.Linux: {},
            OSTypes.MacOS: {}
        }
        self._registerWindowServices()

    def getService(self, serviceType: Type[ServiceBase]) -> Optional[ServiceBase]:
        try:
            rss = self._registeredServices
            if serviceType in rss[OSTypes.Any]:
                return rss[OSTypes.Any][serviceType]

            return rss[self.OSType][serviceType]
        except KeyError:
            return

    def _registerWindowServices(self):
        if self.OSType == OSTypes.Windows:
            from autoinsight.services.WindowKnowledgeService import WindowKnowledgeService
            from wmi import WMI

            serviceMap = self._registeredServices[OSTypes.Windows]
            serviceMap.setdefault(KnowledgeServiceBase, WindowKnowledgeService())
            serviceMap.setdefault(WMI, WMI())



from .ProcessBase import ProcessBase
from .WindowShellBase import WindowShellBase


class Command(WindowShellBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def start(self) -> ProcessBase:
        pass

    def __enter__(self):
        pass

    def __exit__(self):
        pass

    def setCurrent(self):
        pass



from abc import abstractmethod
from typing import Iterable, Optional

from autoinsight.common.models.Point import Point
from autoinsight.common.models.Rectangle import Rectangle
from autoinsight.common.models.Size import Size
from autoinsight.ident.target.TargetBase import TargetBase
from .ProcessBase import ProcessBase


class FormBase(ProcessBase):
    def __init__(self,
                 parent: Optional[ProcessBase] = None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._x = 0
        self._y = 0
        self._width = 0
        self._height = 0
        self._parent: Optional[ProcessBase] = parent

    @property
    def title(self) -> str:
        pass

    @property
    def position(self) -> Point:
        return Point(x=self._x, y=self._y)

    @property
    def parent(self) -> Optional[ProcessBase]:
        return self._parent

    @property
    def targets(self) -> Iterable[TargetBase]:
        pass

    @property
    def size(self) -> Size:
        return Size(width=self._width, height=self._height)

    @property
    def rectangle(self) -> Rectangle:
        return Rectangle(left=self._x, top=self._y, width=self._width, height=self._height)

    @abstractmethod
    def focus(self):
        pass

    @abstractmethod
    def maximize(self):
        pass

    @abstractmethod
    def minimize(self):
        pass

    @abstractmethod
    def drag(self):
        pass

    @abstractmethod
    def drop(self):
        pass

    @abstractmethod
    def dragTo(self, x: int, y: int):
        pass

    @abstractmethod
    def snapshot(self):
        pass

    @abstractmethod
    def scroll(self):
        pass



import yaml
import pathlib
import os

from .ServiceBase import ServiceBase


class ConfigurationService(ServiceBase):
    def __init__(self, *args, **kwargs):
        pass

    def load(self, configFilePath: str):
        pass

    def parse(self, configContent: str):
        pass

    def refreshConfig(self):
        pass

    def save(self):
        pass

    def _validate(self, yamlContent: str):
        """
        Validate configuration content base on the
        https://www.andrewvillazon.com/validate-yaml-python-schema/
        """
        pass

    def merge(self, *configs, **kwargs):
        """
        Merge multiple levels of configurations into a single one. The later will
        overwrite the previous settings.
        """
        pass



from abc import abstractmethod

from autoinsight.common.ObjectBase import ObjectBase


class StepBase(ObjectBase):
    @abstractmethod
    def action(self):
        pass

    @property
    def result(self):
        pass



from .ControlBase import ControlBase


class Button(ControlBase):
    def __init__(self, query: str, *args, **kwargs):
        super().__init__(query, *args, **kwargs)

    def __repr__(self):
        return

    def __str__(self):
        pass

    def click(self) -> bool:
        if super().click():
            return True
        elif self.isToggleable():
            return self.toggle()
        else:
            return False

    def toggle(self) -> bool:
        if self.automationInstance:
            try:
                self.automationInstance.toggle()
                return True
            except:
                return False

    def isToggleable(self) -> bool:
        try:
            if not self.automationInstance:
                return False
            else:
                return not not self.automationInstance.iface_toggle
        except:
            return False



import os
import platform
import signal
from collections.abc import Sequence
from time import sleep
from typing import Iterable, Mapping, Optional

import pyautogui
from wmi import WMI, _wmi_object

from autoinsight.common.EnumTypes import ButtonTypes
from autoinsight.common.models.Point import Point
from autoinsight.services.KnowledgeServiceBase import KnowledgeServiceBase
from autoinsight.services.WindowKnowledgeService import WindowKnowledgeService
from .Command import Command
from .OSBase import OSBase
from .PowerShell import PowerShell
from .ProcessBase import ProcessBase
from .WindowGUIApplication import WindowGUIApplication


class WindowOS(OSBase):
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._wmi: _wmi_object = self._ioc.getService(WMI)
        self._ks: WindowKnowledgeService = self._ioc.getService(KnowledgeServiceBase)

    def __repr__(self):
        if not self._repr:
            self._repr = f"{self.__str__()} id:{self.id} {self.version}"
        return self._repr

    def __str__(self):
        if not self._str:
            self._str = "Window OS Context"
        return self._str

    @property
    def processes(self) -> Iterable[ProcessBase]:
        pass

    @property
    def desktop(self):
        pass

    @property
    def ks(self) -> WindowKnowledgeService:
        return self._ks

    @property
    def wmi(self) -> _wmi_object:
        return self._wmi

    @property
    def userHome(self):
        return self.environ["USERPROFILE"]

    @property
    def environ(self) -> Mapping[str, str]:
        return self._environ

    @property
    def monitors(self):
        pass

    @property
    def systeminfo(self) -> str:
        pass

    @property
    def version(self) -> str:
        return f"{platform.system()} {platform.version()} {platform.release()} {platform.win32_edition()}"

    @property
    def drivers(self) -> Iterable[str]:
        pass

    @property
    def disks(self) -> Iterable[str]:
        pass

    @property
    def SBIOS(self) -> str:
        pass

    @property
    def VBIOS(self) -> str:
        pass

    @property
    def services(self):
        pass

    def shutdown(self, delaySeconds: int = 0):
        delaySeconds: int = int(delaySeconds)
        os.system("shutdown /s /t %s" % delaySeconds)

    def holdAndTypeKeys(self, keys: Iterable[str], holdKeys: Sequence[str]):
        countOfControlKeys: int = len(holdKeys)
        if countOfControlKeys == 0:
            self.typeKeys(keys)
        elif countOfControlKeys == 1:
            with pyautogui.hold(holdKeys[0]):
                self.typeKeys(keys)
        elif countOfControlKeys == 2:
            with pyautogui.hold(holdKeys[0]):
                with pyautogui.hold(holdKeys[1]):
                    self.typeKeys(keys)
        elif countOfControlKeys == 3:
            with pyautogui.hold(holdKeys[0]):
                with pyautogui.hold(holdKeys[1]):
                    with pyautogui.hold(holdKeys[2]):
                        self.typeKeys(keys)

    def changeDriverState(self, driverInfo: str, isEnable: bool):
        pass

    def changeServiceState(self, serviceInfo: str, isEnable: bool):
        pass

    def install(self, path: str):
        """Automatic install an application
        """
        pass

    def typeControlKeys(self, controlKeys: Iterable[str]):
        """Send simulated control keys

        Args:
            controlKeys (str): a string contain the control keys like {ENTER} etc.
        """
        pyautogui.write(controlKeys)

    def hotkeys(self, *keys):
        pyautogui.hotkey(*keys)

    def querySystemEvent(self, eventPattern: str, level):
        pass

    def restart(self, delaySeconds: int = 0):
        delaySeconds: int = int(delaySeconds)
        os.system("shutdown /s /r /t %s" % delaySeconds)

    def sleep(self, delaySeconds: int = 0):
        delaySeconds: int = int(delaySeconds)

    def hibernate(self, delaySeconds: int = 0):
        delaySeconds: int = int(delaySeconds)
        os.system("shutdown /s /hybrid /t %s" % delaySeconds)

    def launchApp(self, cmdline: str, isAsAdmin: bool = False, *args, **kwargs) -> WindowGUIApplication:
        instance = self.ks.recognize(cmdline)
        if instance:
            return WindowGUIApplication(automationInstance=instance)
        else:
            return WindowGUIApplication(cmdline=cmdline, *args, **kwargs)

    def launchShell(self, isAsAdmin: bool = False, *args, **kwargs) -> Optional[Command]:
        return Command(*args, **kwargs)

    def launchPowershell(self, *args, **kwargs) -> Optional[PowerShell]:
        return PowerShell(*args, **kwargs)

    def kill(self, processId: int = 0, processName: str = ""):
        if processId:
            os.kill(processId, signal.SIGKILL)

    def terminate(self, processId: int = 0, processName: str = ""):
        if processId:
            os.kill(processId, signal.SIGTERM)

    def updateDriver(self):
        pass

    def typeKeys(self, visibleKeys: Iterable[str], intervalSec: int = 0.25):
        """
        Simulate typing the visible characters
        """
        pyautogui.write(visibleKeys, interval=intervalSec)

    def snapshot(self):
        pass

    def select(self, start: Point, end: Point, button: ButtonTypes = ButtonTypes.Left):
        pyautogui.mouseDown(x=start.x, y=start.y, button=button.str)
        pyautogui.mouseUp(x=end.x, y=end.y, button=button.str)

    def wait(self, timeoutSeconds: int = 0):
        sleep(timeoutSeconds)

    def switchUser(self):
        pass

    def logout(self):
        pass

    def waitForWindow(self, query: str) -> WindowGUIApplication:
        pass



from abc import ABC
from typing import Optional

from .TargetBase import TargetBase
from autoinsight.common.EnumTypes import ScrollDirectionTypes
from autoinsight.ident.IdentObjectBase import IdentObjectBase


class ControlBase(TargetBase, ABC):
    @property
    def next(self) -> Optional[IdentObjectBase]:
        pass

    @property
    def previous(self) -> Optional[IdentObjectBase]:
        pass

    def focus(self) -> bool:
        if self.automationInstance:
            try:
                self.automationInstance.set_focus()
                return True
            except:
                return False

    def scroll(self, direction: ScrollDirectionTypes = ScrollDirectionTypes.Down):
        if isinstance(direction, str):
            direction: ScrollDirectionTypes = ScrollDirectionTypes.fromStr(direction)

        if self.isScrollable():
            self.automationInstance.scroll(direction.Down.str, "line")



from typing import Iterable, Optional

from .GUIApplicationBase import GUIApplicationBase
from .ProcessBase import ProcessBase
from .WindowForm import WindowForm


class WindowGUIApplication(GUIApplicationBase):
    def __init__(self, *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def forms(self) -> Optional[Iterable[WindowForm]]:
        return

    @property
    def currentForm(self) -> Optional[WindowForm]:
        return self._currentForm

    def focus(self) -> bool:
        if self.automationInstance:
            try:
                self.automationInstance.set_focus()
                return True
            except:
                return False

    def start(self) -> ProcessBase:
        pass

    def wait(self, timeoutSeconds: int = 0):
        pass

    def snapshot(self):
        pass



from abc import abstractmethod
from typing import Tuple, Optional

from PIL import Image
from numpy import array
from requests import post, get, Response

from autoinsight.common.models.BoundingBox import BoundingBox
from .ServiceBase import ServiceBase


class OCRServiceBase(ServiceBase):

    @classmethod
    def _getImageFromURL(cls, auth, imageUrl, payload):
        if payload:
            result: Response = post(imageUrl, data=payload, auth=auth)
        else:
            result: Response = get(imageUrl, auth=auth)
        ok = 200
        if result.status_code == ok:
            mode = result.headers["mode"]
            width = int(result.headers["width"])
            height = int(result.headers["height"])
            image = Image.frombuffer(mode, (width, height), result.content)
        return image

    @abstractmethod
    def getTextFromImageFile(self, imagePath) -> str:
        pass

    @abstractmethod
    def getTextFromImageBinary(self, imageBinary: bytes, mode: str, size: Tuple[int, int]) -> str:
        pass

    @abstractmethod
    def getTextFromImage(self, image: Image) -> str:
        pass

    @abstractmethod
    def getTextFromImageUrl(self, imageUrl: str, payload: str, auth: Optional[Tuple[str, str]] = None) -> str:
        pass

    @abstractmethod
    def getTextFromImageArray(self, imageArray: array) -> str:
        pass

    @abstractmethod
    def getBoxFromImageFile(self, imagePath) -> Optional[list[BoundingBox]]:
        pass

    @abstractmethod
    def getBoxFromImageBinary(self, imageBinary: bytes, mode: str, size: Tuple[int, int]) -> Optional[list[BoundingBox]]:
        pass

    @abstractmethod
    def getBoxFromImage(self, image: Image) -> Optional[list[BoundingBox]]:
        pass

    @abstractmethod
    def getBoxFromImageUrl(self, imageUrl: str,
                           payload: str = None,
                           auth: Optional[Tuple[str, str]] = None) -> Optional[list[BoundingBox]]:
        pass

    @abstractmethod
    def getBoxFromImageArray(self, imageArray: array) -> Optional[list[BoundingBox]]:
        pass



from abc import abstractmethod

from autoinsight.common.ObjectBase import ObjectBase
from autoinsight.services.ConfigurationService import ConfigurationService
from autoinsight.services.ContextManagementService import ContextManagementService
from autoinsight.services.IoCService import IoCService


class ScriptBase(ObjectBase):
    @abstractmethod
    def __init__(self,
                 ioc: IoCService = IoCService(),
                 *args,
                 **kwargs):
        self._ioc: IoCService = ioc
        self._cms: ContextManagementService = ioc.getService(ContextManagementService)
        self._cs: ConfigurationService = ioc.getService(ConfigurationService)

    @property
    def steps(self):
        pass

    @property
    def config(self):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def resume(self):
        pass

    @abstractmethod
    def pause(self):
        pass

    @abstractmethod
    def save(self, path: str) -> bool:
        pass



from time import sleep
from functools import wraps

from .DecoratorBase import DecoratorBase


# TODO add delay to the actions such as 1 second for each action
class Delay(DecoratorBase):
    def __init__(self, func, *args, **kwargs):
        pass





from typing import Callable, Optional, Iterable

from .ControlBase import ControlBase


class ComboBox(ControlBase):
    def __init__(self, query: str, *args, **kwargs):
        super().__init__(query, *args, **kwargs)

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        pass

    @property
    def values(self) -> Iterable[str]:
        pass

    def iterateValues(self, intervalSecond: int = 2,
                      callbackOnEachValue: Optional[Callable[[str], bool]] = None):
        pass

    def select(self, value: Optional[str] = None, index: int = -1):
        pass



from abc import abstractmethod
from uuid import UUID

from autoinsight.common.ObjectBase import ObjectBase
from autoinsight.common.Utils import GUID


class IdentObjectBase(ObjectBase):
    def __init__(self, *args, **kwargs):
        self._id = GUID()
        self._repr = None
        self._str = None
        self._automationInstance = None

    @property
    def id(self) -> UUID:
        return self._id

    @property
    def description(self) -> str:
        """
        A description that will unique locate the object within the context
        """
        pass

    @property
    def classname(self):
        """
        The control class name
        """
        return type(self).__name__

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def snapshot(self):
        pass



import os
import sys

from .ScriptBase import ScriptBase


class Script(ScriptBase):
    """
    Should have the ability to persistent the execution state after restart
    """

    def run(self):
        pass

    def load(self):
        pass

    def resume(self):
        pass

    def pause(self):
        pass

    def __init__(self, location: str = os.path.dirname(sys.argv[0]), *args, **kwarg):
        self._location = location

    @property
    def location(self) -> str:
        return self._location



from typing import Iterable


from .FormBase import FormBase
from ..target.TargetBase import TargetBase


class WindowForm(FormBase):
    def __int__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        pass

    def __str__(self):
        pass

    @property
    def targets(self) -> Iterable[TargetBase]:
        pass

    @property
    def title(self) -> str:
        pass

    def snapshot(self):
        pass

    def scroll(self):
        pass

    def wait(self, timeoutSeconds: int = 0):
        pass

    def focus(self):
        pass

    def maximize(self):
        pass

    def minimize(self):
        pass

    def drag(self):
        pass

    def drop(self):
        pass

    def dragTo(self, x: int, y: int):
        pass



from dataclasses import dataclass

from .Point import Point
from .ShapeBase import ShapeBase
from .Size import Size


@dataclass
class Rectangle(ShapeBase):
    width: int
    height: int

    @property
    def center(self) -> Point:
        return Point(x=self.left + self.width // 2, y=self.top + self.height // 2)

    @property
    def size(self) -> Size:
        return Size(width=self.width, height=self.height)



import collections
from typing import Union, NewType, Optional

from pywinauto.application import WindowSpecification
from pywinauto.uia_element_info import UIAElementInfo
# TODO the imports should be replaced with the wrapper in autoinsght.automation.*
from pywinauto.win32_element_info import HwndElementInfo

# TODO in the future we need to separate the implementation for technical instance and visual instance
AutomationInstance = NewType("AutomationInstance", Optional[Union[UIAElementInfo,
                                                                  HwndElementInfo,
                                                                  WindowSpecification]])

ElementTreeNode = collections.namedtuple("ElementTreeNode", ["elem", "id", "children"])
ElementsInfo = collections.namedtuple("ElementsInfo", ["ctrlTreeRoot",
                                                       "textCtrls",
                                                       "allCtrlIndexNameMaps",
                                                       "allCtrl"])



from __future__ import annotations

import os
import signal
from abc import abstractmethod
from typing import Optional

from autoinsight.ident.AutomationTyping import AutomationInstance
from .ContextBase import ContextBase


class ProcessBase(ContextBase):
    def __init__(self,
                 processId: Optional[int] = 0,
                 cmdline: Optional[str] = None,
                 title: Optional[str] = None,
                 workdir: Optional[str] = None,
                 automationInstance: Optional[AutomationInstance] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._processId = processId
        self._cmdline = cmdline
        self._title = title
        self._workdir = workdir
        self._exitcode = 0
        self._automationInstance: Optional[AutomationInstance] = automationInstance

    def __repr__(self):
        if not self._repr:
            self._repr = f"{self.__str__()} id:{self.processId} cmdline:{self.cmdline} workdir:{self.workdir}"
        return self._repr

    def __str__(self):
        if not self._str:
            if self.cmdline:
                self._str = os.path.basename(self.cmdline)
            elif self._title:
                self._str = self._title

        return self._str

    @property
    def automationInstance(self) -> Optional[AutomationInstance]:
        if not self._automationInstance:
            self._automationInstance = self._cms.os.find(
                f"{self._processId} {self.cmdline} {self._title} {self.workdir}")
        return self._automationInstance

    @property
    def exitcode(self) -> int:
        return self._exitcode

    @property
    def workdir(self) -> str:
        return self._workdir

    @property
    def cmdline(self) -> str:
        return self._cmdline

    @property
    def processId(self) -> int:
        return self._processId

    @abstractmethod
    def start(self) -> ProcessBase:
        pass

    def __enter__(self):
        super().__enter__()
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def new(cls, *args, **kwargs) -> ProcessBase:
        pass

    def close(self):
        if self._processId:
            os.kill(self._processId, signal.SIGTERM)
            self.tearDown()
            return

        if self._automationInstance:
            self._automationInstance.close()
            self.tearDown()
            return

    def kill(self):
        if self._processId:
            os.kill(self._processId, signal.SIGKILL)
            self.tearDown()
            return

    def _printElementsTree(self):
        """
        Print debug element tree information to the console.
        """
        pass



from typing import Optional, Any
from enum import Enum, EnumMeta


def _fromStr(cls: EnumMeta, value: str, default: Any = None) -> Optional[Enum]:
    for item in cls:
        if item.value == value:
            return item

    return default


def _str(self: Enum) -> str:
    return str(self.value)


Enum.fromStr = classmethod(_fromStr)
Enum.str = property(_str)



from abc import abstractmethod
from collections.abc import Sequence
from os import environ
from typing import Iterable, Mapping, Optional

from autoinsight.common.EnumTypes import ButtonTypes
from autoinsight.common.models.Point import Point
from .ContextBase import ContextBase
from .GUIApplicationBase import GUIApplicationBase
from .ProcessBase import ProcessBase
from .ShellBase import ShellBase
from ..AutomationTyping import AutomationInstance


class OSBase(ContextBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._environ: Mapping[str] = environ

    @property
    def userHome(self):
        pass

    @property
    def processes(self) -> Iterable[ProcessBase]:
        pass

    @property
    def desktop(self):
        """
        Display desktop and minimize all the other application
        """
        pass

    @property
    def environ(self) -> Mapping[str, str]:
        """
        Return the system environment variable mapping
        """
        return self._environ

    @property
    def monitors(self):
        pass

    @property
    def systeminfo(self) -> str:
        pass

    @property
    def version(self) -> str:
        pass

    @property
    def SBIOS(self) -> str:
        pass

    @property
    def VBIOS(self) -> str:
        pass

    @property
    def drivers(self) -> Iterable[str]:
        """
        Return the hardware driver information
        """
        pass

    @abstractmethod
    def launchApp(self, path, isAsAdmin: bool = False) -> GUIApplicationBase:
        pass

    @abstractmethod
    def launchShell(self, isAsAdmin: bool = False, *args, **kwargs) -> ShellBase:
        pass

    @abstractmethod
    def kill(self, processId: int = 0, processName: str = ""):
        pass

    @abstractmethod
    def terminate(self, processId: int = 0, processName: str = ""):
        pass

    @abstractmethod
    def disks(self) -> Iterable[str]:
        """
        Return all the available hard drive partitions
        """
        pass

    @abstractmethod
    def changeDriverState(self, driverInfo: str, isEnable: bool):
        pass

    @abstractmethod
    def updateDriver(self):
        pass

    @abstractmethod
    def install(self, path: str):
        """
        Automatic an application
        """
        pass

    @abstractmethod
    def typeKeys(self, visibleKeys: Iterable[str], intervalSec: int = 0.25):
        """
        Simulate typing the visible characters
        """
        pass

    @abstractmethod
    def typeControlKeys(self, controlKeys: Iterable[str]):
        pass

    @abstractmethod
    def hotkeys(self, *keys):
        pass

    @abstractmethod
    def holdAndTypeKeys(self, keys: Iterable[str], holdKeys: Sequence[str]):
        """
        Hold the control keys specified in holdKeys and press the keys.
        """
        pass

    @abstractmethod
    def snapshot(self):
        pass

    @abstractmethod
    def querySystemEvent(self, eventPattern: str, level):
        pass

    @abstractmethod
    def select(self, start: Point, end: Point, button: ButtonTypes = ButtonTypes.Left):
        pass

    @abstractmethod
    def restart(self, delaySeconds: int = 0):
        pass

    @abstractmethod
    def sleep(self, delaySeconds: int = 0):
        pass

    @abstractmethod
    def hibernate(self, delaySeconds: int = 0):
        pass

    @abstractmethod
    def shutdown(self, delaySeconds: int = 0):
        pass

    @abstractmethod
    def switchUser(self):
        pass

    @abstractmethod
    def logout(self):
        pass

    def find(self, query: str, *args, **kwargs) -> Optional[AutomationInstance]:
        if self._cms.currentContext != self:
            # TODO improve the find logic
            return self._cms.currentContext



from enum import Enum

import autoinsight.extend.ExtendEnum  # noqa:F401


class OSTypes(Enum):
    Any = "Any"
    Windows = "Windows"
    Linux = "Linux"
    Java = "Java"
    MacOS = "Darwin"


class ButtonTypes(Enum):
    Left = "left"
    Middle = "middle"
    Right = "right"


class ScrollDirectionTypes(Enum):
    Up = "up"
    Down = "down"
    Left = "left"
    Right = "right"
